@@ -1,12 +1,15 @@
 """Custom client for interacting with Spotify's Web API"""
-from datetime import datetime, timedelta, date  # pylint: disable=unused-import
+from __future__ import annotations
+
+from datetime import date, datetime, timedelta
 from enum import Enum
 from json import dumps
-from logging import getLogger, DEBUG
+from logging import DEBUG, getLogger
 from re import sub
+from typing import Any, Callable, Dict, List, Literal, Optional, TypedDict, Union
 
-from requests import get, Response, post
-from spotipy import SpotifyOAuth, CacheFileHandler
+from requests import Response, get, post
+from spotipy import CacheFileHandler, SpotifyOAuth
 
 from wg_utilities.functions import chunk_list
 from wg_utilities.loggers import add_stream_handler
@@ -24,6 +27,89 @@ class AlbumType(Enum):
     COMPILATION = "compilation"
 
 
+class _SpotifyEntityInfo(TypedDict):
+    description: str
+    href: str
+    id: str
+    name: str
+    uri: str
+
+
+class _AlbumTracksItemInfo(TypedDict):
+    href: str
+    items: List[_TrackInfo]  # type: ignore
+    limit: int
+    next: Union[str, None]
+    offset: int
+    previous: Union[str, None]
+    total: int
+
+
+class _AlbumInfo(_SpotifyEntityInfo):
+    album_type: Literal["album", "single", "compilation"]
+    artists: List[_ArtistInfo]
+    available_markets: List[str]
+    external_urls: Dict[Literal["spotify"], str]
+    images: List[Dict[str, Union[str, int]]]
+    release_date: str
+    release_date_precision: Literal["year", "month", "day", None]
+    restrictions: Dict[str, str]
+    total_tracks: int
+    tracks: _AlbumTracksItemInfo
+    type: Literal["album"]
+
+
+class _ArtistInfo(_SpotifyEntityInfo):
+    external_urls: Dict[Literal["spotify"], str]
+    followers: Dict[str, Union[Optional[str], int]]
+    genres: List[str]
+    images: List[Dict[str, Union[str, int]]]
+    popularity: int
+    type: Literal["artist"]
+
+
+class _PlaylistInfo(_SpotifyEntityInfo):
+    collaborative: bool
+    external_urls: Dict[Literal["spotify"], str]
+    followers: Dict[str, Union[Optional[str], int]]
+    images: List[Dict[str, Union[str, int]]]
+    owner: _UserInfo
+    public: bool
+    snapshot_id: str
+    tracks: List[_TrackInfo]  # type: ignore
+    type: Literal["playlist"]
+
+
+class _TrackInfo(_SpotifyEntityInfo):
+    album: _AlbumInfo
+    artists: List[_ArtistInfo]
+
+
+class _UserInfo(_SpotifyEntityInfo):
+    display_name: str
+
+
+class _TrackAudioFeaturesInfo(TypedDict):
+    acousticness: float
+    analysis_url: str
+    danceability: float
+    duration_ms: int
+    energy: float
+    id: str
+    instrumentalness: float
+    key: int
+    liveness: float
+    loudness: float
+    mode: int
+    speechiness: float
+    tempo: float
+    time_signature: int
+    track_href: str
+    type: Literal["audio_features"]
+    uri: str
+    valence: float
+
+
 class SpotifyEntity:
     """Parent class for all Spotify entities (albums, artists, etc.)
 
@@ -35,13 +121,18 @@ class SpotifyEntity:
         metadata (dict): any extra metadata about this entity
     """
 
-    def __init__(self, json, spotify_client=None, metadata=None):
+    def __init__(
+        self,
+        json: _SpotifyEntityInfo,
+        spotify_client: SpotifyClient,
+        metadata: Optional[Dict[str, Any]] = None,
+    ):
         self.json = json
         self._spotify_client = spotify_client
         self.metadata = metadata or {}
 
     @property
-    def pretty_json(self):
+    def pretty_json(self) -> str:
         """
         Returns:
             str: a "pretty" version of the JSON, used for debugging etc.
@@ -49,7 +140,7 @@ class SpotifyEntity:
         return dumps(self.json, indent=4, default=str)
 
     @property
-    def description(self):
+    def description(self) -> Optional[str]:
         """
         Returns:
             str: the description of the entity
@@ -57,7 +148,7 @@ class SpotifyEntity:
         return self.json.get("description")
 
     @property
-    def endpoint(self):
+    def endpoint(self) -> Optional[str]:
         """
         Returns:
             str: A link to the Web API endpoint providing full details of the entity
@@ -65,7 +156,7 @@ class SpotifyEntity:
         return self.json.get("href")
 
     @property
-    def id(self):
+    def id(self) -> Optional[str]:
         """
         Returns:
             str: The base-62 identifier for the entity
@@ -73,15 +164,15 @@ class SpotifyEntity:
         return self.json.get("id")
 
     @property
-    def name(self):
+    def name(self) -> str:
         """
         Returns:
             str: the name of the entity
         """
-        return self.json.get("name")
+        return self.json["name"]
 
     @property
-    def uri(self):
+    def uri(self) -> Optional[str]:
         """
         Returns:
             str: the Spotify URI of this entity
@@ -89,63 +180,84 @@ class SpotifyEntity:
 
         return self.json.get("uri", f"spotify:{type(self).__name__.lower()}:{self.id}")
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, SpotifyEntity):
+            return NotImplemented
         return self.id == other.id
 
-    def __gt__(self, other):
+    def __gt__(self, other: object) -> bool:
+        if not isinstance(other, SpotifyEntity):
+            return NotImplemented
         return self.name.lower() > other.name.lower()
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash(repr(self))
 
-    def __lt__(self, other):
+    def __lt__(self, other: SpotifyEntity) -> bool:
         return self.name.lower() < other.name.lower()
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f'{type(self).__name__}(id="{self.id}")'
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"{self.name} ({self.id})"
 
 
 class User(SpotifyEntity):
     """A Spotify user, usually just the current user"""
 
+    json: _UserInfo
+
+    def __init__(
+        self,
+        json: _UserInfo,
+        spotify_client: SpotifyClient,
+        metadata: Optional[Dict[str, Any]] = None,
+    ):
+        super().__init__(json=json, spotify_client=spotify_client, metadata=metadata)
+
     @property
-    def name(self):
+    def name(self) -> str:
         """
         Returns:
             str: the display name of the User
         """
 
-        return self.json.get("display_name")
+        return self.json["display_name"]
 
 
 class Track(SpotifyEntity):
     """A track on Spotify"""
 
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self._artists = None
-        self._audio_features = None
+    json: _TrackInfo
+
+    def __init__(
+        self,
+        json: _TrackInfo,
+        spotify_client: SpotifyClient,
+        metadata: Optional[Dict[str, Any]] = None,
+    ):
+        super().__init__(json=json, spotify_client=spotify_client, metadata=metadata)
+        self._artists: Optional[List[Artist]] = None
+        self._audio_features: Optional[_TrackAudioFeaturesInfo] = None
 
     @property
-    def album(self):
+    def album(self) -> Album:
         """
         Returns:
             Album: the album which this track is from
         """
 
-        return Album(self.json.get("album", {}), self._spotify_client)
+        return Album(self.json["album"], self._spotify_client)
 
     @property
-    def artists(self):
+    def artists(self) -> List[Artist]:
         """
         Returns:
             list(Artist): a list of the artists who contributed to this track
         """
 
-        if not self._artists:
+        if self._artists is None:
             self._artists = [
                 Artist(item, self._spotify_client)
                 for item in self.json.get("artists", [])
@@ -154,12 +266,12 @@ class Track(SpotifyEntity):
         return self._artists
 
     @property
-    def audio_features(self):
+    def audio_features(self) -> _TrackAudioFeaturesInfo:
         """
         Returns:
             dict: the JSON response from the Spotify /audio-features endpoint
         """
-        if not self._audio_features:
+        if self._audio_features is None:
             self._audio_features = self._spotify_client.get_json_response(
                 f"/audio-features/{self.id}"
             )
@@ -167,7 +279,7 @@ class Track(SpotifyEntity):
         return self._audio_features
 
     @property
-    def release_date(self):
+    def release_date(self) -> Union[date, str, None]:
         """
         Returns:
             date: the date the track's album was first released
@@ -175,7 +287,7 @@ class Track(SpotifyEntity):
         return self.album.release_date
 
     @property
-    def tempo(self):
+    def tempo(self) -> Union[float, None]:
         """
         Returns:
             float: the tempo of the track, in BPM
@@ -186,19 +298,19 @@ class Track(SpotifyEntity):
 class Artist(SpotifyEntity):
     """An artist on Spotify"""
 
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self._albums = None
+    def __init__(self, json: _ArtistInfo, spotify_client: SpotifyClient):
+        super().__init__(json=json, spotify_client=spotify_client)
+        self._albums: Optional[List[Album]] = None
 
     @property
-    def albums(self):
+    def albums(self) -> List[Album]:
         """
         Returns:
             list: A list of albums this artist has contributed to
         """
         if not self._albums:
             self._albums = [
-                Album(item, self._spotify_client)
+                Album(item, self._spotify_client)  # type: ignore[arg-type]
                 for item in self._spotify_client.get_items_from_url(
                     f"/artists/{self.id}/albums"
                 )
@@ -210,19 +322,21 @@ class Artist(SpotifyEntity):
 class Album(SpotifyEntity):
     """An album on Spotify"""
 
-    def __init__(self, json, spotify_client=None):
+    json: _AlbumInfo
+
+    def __init__(self, json: _AlbumInfo, spotify_client: SpotifyClient):
         super().__init__(json, spotify_client)
-        self._artists = None
-        self._tracks = None
+        self._artists: Optional[List[Artist]] = None
+        self._tracks: Optional[List[Track]] = None
 
     @property
-    def artists(self):
+    def artists(self) -> List[Artist]:
         """
         Returns:
-            list: a list of artists that appear on this album
+            list(Artist): a list of the artists who contributed to this track
         """
 
-        if not self._artists:
+        if self._artists is None:
             self._artists = [
                 Artist(item, self._spotify_client)
                 for item in self.json.get("artists", [])
@@ -231,7 +345,7 @@ class Album(SpotifyEntity):
         return self._artists
 
     @property
-    def release_date(self):
+    def release_date(self) -> Union[date, str, None]:
         """
         Returns:
             date: the date the album was first released
@@ -246,7 +360,7 @@ class Album(SpotifyEntity):
         return release_date_str
 
     @property
-    def release_date_precision(self):
+    def release_date_precision(self) -> Literal["year", "month", "day", None]:
         """
         Returns:
             str: the precision with which release_date value is known
@@ -254,7 +368,7 @@ class Album(SpotifyEntity):
         return self.json.get("release_date_precision")
 
     @property
-    def tracks(self):
+    def tracks(self) -> List[Track]:
         """
         Returns:
             list: a list of tracks on this album
@@ -266,16 +380,16 @@ class Album(SpotifyEntity):
                 for item in self.json.get("tracks", {}).get("items", [])
             ]
 
-            if next_url := self.json.get("next"):
+            if next_url := self.json.get("tracks", {}).get("next"):
                 self._tracks.extend(
-                    Track(item, self._spotify_client)
+                    Track(item, self._spotify_client)  # type: ignore[arg-type]
                     for item in self._spotify_client.get_items_from_url(next_url)
                 )
 
         return self._tracks
 
     @property
-    def type(self):
+    def type(self) -> AlbumType:
         """
         Returns:
             AlbumType: the type of album this is
@@ -287,20 +401,30 @@ class Album(SpotifyEntity):
 class Playlist(SpotifyEntity):
     """A Spotify playlist"""
 
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self._tracks = None
+    json: _PlaylistInfo
+
+    def __init__(
+        self,
+        json: _PlaylistInfo,
+        spotify_client: SpotifyClient,
+        metadata: Optional[Dict[str, Any]] = None,
+    ):
+        super().__init__(json=json, spotify_client=spotify_client, metadata=metadata)
+        self._tracks: Optional[List[Track]] = None
 
     @property
-    def tracks(self):
+    def tracks(self) -> List[Track]:
         """
         Returns:
             list: a list of tracks in this playlist
         """
 
-        if not self._tracks:
+        if self._tracks is None:
             self._tracks = [
-                Track(item.get("track", {}), self._spotify_client)
+                Track(
+                    item.get("track", {}),  # type: ignore[arg-type,union-attr]
+                    self._spotify_client,
+                )
                 for item in self._spotify_client.get_items_from_url(
                     f"/playlists/{self.id}/tracks"
                 )
@@ -309,18 +433,21 @@ class Playlist(SpotifyEntity):
         return self._tracks
 
     @property
-    def owner(self):
+    def owner(self) -> User:
         """
         Returns:
             User: the Spotify user who owns this playlist
         """
 
-        return User(self.json.get("owner", {}))
+        return User(self.json["owner"], self._spotify_client)
 
-    def __contains__(self, track):
+    def __contains__(self, track: Track) -> bool:
         return track.id in [track.id for track in self.tracks]
 
-    def __gt__(self, other):
+    def __gt__(self, other: object) -> bool:
+        if not isinstance(other, Playlist):
+            return NotImplemented
+
         if self.name.lower() == other.name.lower():
             if self.owner.id == self._spotify_client.current_user.id:
                 return False
@@ -330,7 +457,10 @@ class Playlist(SpotifyEntity):
 
         return self.name.lower() > other.name.lower()
 
-    def __lt__(self, other):
+    def __lt__(self, other: object) -> bool:
+        if not isinstance(other, Playlist):
+            return NotImplemented
+
         if self.name.lower() == other.name.lower():
 
             if self.owner.id == self._spotify_client.current_user.id:
@@ -341,7 +471,7 @@ class Playlist(SpotifyEntity):
 
         return self.name.lower() < other.name.lower()
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"{self.name} ({self.id}) - owned by {self.owner}"
 
 
@@ -389,13 +519,13 @@ class SpotifyClient:
     def __init__(
         self,
         *,
-        client_id=None,
-        client_secret=None,
-        redirect_uri="http://localhost:8080",
-        scope=None,
-        oauth_manager=None,
-        log_requests=False,
-        creds_cache_path=None,
+        client_id: Optional[str] = None,
+        client_secret: Optional[str] = None,
+        redirect_uri: str = "http://localhost:8080",
+        scope: Optional[Union[str, List[str]]] = None,
+        oauth_manager: Optional[SpotifyOAuth] = None,
+        log_requests: bool = False,
+        creds_cache_path: Optional[str] = None,
     ):
         self.oauth_manager = oauth_manager or SpotifyOAuth(
             client_id=client_id,
@@ -407,14 +537,20 @@ class SpotifyClient:
         self.log_requests = log_requests
         self.api_call_count = 0
 
-        self._current_user = None
+        self._current_user: Optional[User] = None
 
-        self._albums = None
-        self._playlists = None
-        self._tracks = None
+        self._albums: Optional[List[Album]] = None
+        self._playlists: Optional[List[Playlist]] = None
+        self._tracks: Optional[List[Track]] = None
 
-    def _get(self, url, params=None):
-        """Wrapper for GET requests which covers authentication, URL parsing, etc etc
+    def _get(
+        self,
+        url: str,
+        params: Optional[
+            Dict[str, Union[str, float, int, bool, Dict[str, Any]]]
+        ] = None,
+    ) -> Response:
+        """Wrapper for GET requests which covers authentication, URL parsing, etc. etc.
 
         Args:
             url (str): the URL path to the endpoint (not necessarily including the
@@ -447,8 +583,14 @@ class SpotifyClient:
 
         return res
 
-    def _post(self, url, json=None):
-        """Wrapper for POST requests which covers authentication, URL parsing, etc etc
+    def _post(
+        self,
+        url: str,
+        json: Optional[
+            Dict[str, Union[str, int, float, bool, List[str], Dict[Any, Any]]]
+        ] = None,
+    ) -> Response:
+        """Wrapper for POST requests which covers authentication, URL parsing, etc. etc.
 
         Args:
             url (str): the URL path to the endpoint (not necessarily including the
@@ -481,8 +623,22 @@ class SpotifyClient:
         return res
 
     def get_items_from_url(
-        self, url, params=None, *, hard_limit=1000000, limit_func=None
-    ):
+        self,
+        url: str,
+        params: Optional[
+            Dict[str, Union[str, int, float, bool, Dict[str, Any]]]
+        ] = None,
+        *,
+        hard_limit: int = 1000000,
+        limit_func: Optional[Callable[[Dict[str, Any]], bool]] = None,
+    ) -> List[
+        Union[
+            _PlaylistInfo,
+            _TrackInfo,
+            _AlbumInfo,
+            List[Dict[Literal["album"], _AlbumInfo]],
+        ]
+    ]:
         """Retrieve a list of items from a given URL, including pagination
 
         Args:
@@ -501,10 +657,17 @@ class SpotifyClient:
         params = params or {}
         params["limit"] = min(50, hard_limit)
 
-        items = []
+        items: List[
+            Union[
+                _PlaylistInfo,
+                _TrackInfo,
+                _AlbumInfo,
+                List[Dict[Literal["album"], _AlbumInfo]],
+            ]
+        ] = []
         res = Response()
         # pylint: disable=protected-access
-        res._content = dumps({"next": url}).encode()  # noqa
+        res._content = dumps({"next": url}).encode()
 
         while (next_url := res.json().get("next")) and len(items) < hard_limit:
             params["limit"] = min(50, hard_limit - len(items))
@@ -523,7 +686,7 @@ class SpotifyClient:
 
         return items
 
-    def get_json_response(self, url):
+    def get_json_response(self, url: str) -> Union[_TrackAudioFeaturesInfo]:
         """Gets a simple JSON object from a URL
 
         Args:
@@ -532,18 +695,18 @@ class SpotifyClient:
         Returns:
             dict: the JSON from the response
         """
-        return self._get(url).json()
+        return self._get(url).json()  # type: ignore
 
     @property
-    def access_token(self):
+    def access_token(self) -> str:
         """
         Returns:
             str: the web API access token
         """
-        return self.oauth_manager.get_access_token(as_dict=False)
+        return str(self.oauth_manager.get_access_token(as_dict=False))
 
     @property
-    def albums(self):
+    def albums(self) -> List[Album]:
         """
         Returns:
             list: a list of albums owned by the current user
@@ -551,14 +714,14 @@ class SpotifyClient:
 
         if not self._albums:
             self._albums = [
-                Album(item.get("album", {}), self)
+                Album(item["album"], self)  # type: ignore
                 for item in self.get_items_from_url("/me/albums")
             ]
 
         return self._albums
 
     @property
-    def playlists(self):
+    def playlists(self) -> List[Playlist]:
         """
         Returns:
             list: a list of playlists owned by the current user
@@ -566,14 +729,14 @@ class SpotifyClient:
 
         if not self._playlists:
             self._playlists = [
-                Playlist(item, self)
+                Playlist(item, self)  # type: ignore[arg-type]
                 for item in self.get_items_from_url("/me/playlists")
             ]
 
         return self._playlists
 
     @property
-    def tracks(self):
+    def tracks(self) -> List[Track]:
         """
         Returns:
             list: a list of tracks owned by the current user
@@ -581,25 +744,27 @@ class SpotifyClient:
 
         if not self._tracks:
             self._tracks = [
-                Track(item.get("track", {}), self)
+                Track(item["track"], self)  # type: ignore
                 for item in self.get_items_from_url("/me/tracks")
             ]
 
         return self._tracks
 
     @property
-    def current_user(self):
+    def current_user(self) -> User:
         """Gets the current user's info
 
         Returns:
             User: an instance of the current Spotify user
         """
         if not self._current_user:
-            self._current_user = User(self._get(f"{self.BASE_URL}/me").json())
+            self._current_user = User(self._get(f"{self.BASE_URL}/me").json(), self)
 
         return self._current_user
 
-    def add_tracks_to_playlist(self, tracks, playlist, *, log_responses=False):
+    def add_tracks_to_playlist(
+        self, tracks: List[Track], playlist: Playlist, *, log_responses: bool = False
+    ) -> None:
         """Add one or more tracks to a playlist
 
         Args:
@@ -618,7 +783,13 @@ class SpotifyClient:
             if log_responses:
                 LOGGER.debug(dumps(res.json()))
 
-    def create_playlist(self, name, description="", public=False, collaborative=False):
+    def create_playlist(
+        self,
+        name: str,
+        description: str = "",
+        public: bool = False,
+        collaborative: bool = False,
+    ) -> Playlist:
         """Create a new playlist under the current user's account
 
         Args:
@@ -644,7 +815,9 @@ class SpotifyClient:
 
         return Playlist(res.json(), self)
 
-    def get_playlists_by_name(self, name, return_all=False):
+    def get_playlists_by_name(
+        self, name: str, return_all: bool = False
+    ) -> Union[List[Playlist], Playlist, None]:
         """Gets Playlist instance(s) which have the given name
 
         Args:
@@ -671,7 +844,7 @@ class SpotifyClient:
 
         return None
 
-    def get_album_by_id(self, id_):
+    def get_album_by_id(self, id_: str) -> Album:
         """Get an album from Spotify based on the ID
 
         Args:
@@ -683,7 +856,7 @@ class SpotifyClient:
 
         return Album(self._get(f"/albums/{id_}").json(), self)
 
-    def get_artist_by_id(self, id_):
+    def get_artist_by_id(self, id_: str) -> Artist:
         """Get an artist from Spotify based on the ID
 
         Args:
@@ -695,7 +868,7 @@ class SpotifyClient:
 
         return Artist(self._get(f"/artists/{id_}").json(), self)
 
-    def get_playlist_by_id(self, id_):
+    def get_playlist_by_id(self, id_: str) -> Playlist:
         """Get a playlist from Spotify based on the ID
 
         Args:
@@ -712,7 +885,7 @@ class SpotifyClient:
 
         return Playlist(self._get(f"/playlists/{id_}").json(), self)
 
-    def get_track_by_id(self, id_):
+    def get_track_by_id(self, id_: str) -> Track:
         """Get a track from Spotify based on the ID
 
         Args:
@@ -724,30 +897,37 @@ class SpotifyClient:
 
         return Track(self._get(f"/tracks/{id_}").json(), self)
 
-    def get_recently_liked_tracks(self, track_limit=100, *, day_limit=None):
+    def get_recently_liked_tracks(
+        self, track_limit: int = 100, *, day_limit: Optional[float] = None
+    ) -> List[Track]:
         """Gets a list of songs which were liked by the current user in the past N days
 
         Args:
             track_limit (int): the number of tracks to return
-            day_limit (int): the number of days (N) to go back in time for
+            day_limit (float): the number of days (N) to go back in time for
 
         Returns:
             list: a list of Track instances
         """
 
-        kwargs = {"hard_limit": track_limit}
+        kwargs: Dict[str, Union[int, Callable[[Any], bool]]] = {
+            "hard_limit": track_limit
+        }
 
-        if day_limit is not None:
-            kwargs["limit_func"] = lambda item: datetime.strptime(
-                item["added_at"], self.DATETIME_FORMAT
-            ) < datetime.utcnow() - timedelta(days=day_limit)
+        if isinstance(day_limit, (float, int)):
+            kwargs["limit_func"] = lambda item: bool(
+                datetime.strptime(item["added_at"], self.DATETIME_FORMAT)
+                # pylint: disable=line-too-long
+                < (datetime.utcnow() - timedelta(days=day_limit))  # type: ignore[arg-type]
+            )
 
         return [
-            Track(item.get("track", {}), self, metadata={"liked_at": item["added_at"]})
-            for item in self.get_items_from_url("/me/tracks", **kwargs)
+            # pylint: disable=line-too-long
+            Track(item["track"], self, metadata={"liked_at": item["added_at"]})  # type: ignore
+            for item in self.get_items_from_url("/me/tracks", **kwargs)  # type: ignore
         ]
 
-    def reset_properties(self):
+    def reset_properties(self) -> None:
         """Resets all list properties"""
 
         self._current_user = None
